from translate_package import (
    pl,
    evaluate,
    LoraConfig,
    TaskType,
    torch,
    get_linear_schedule_with_warmup,
    wandb,
    get_peft_model,
    T5ForConditionalGeneration,
    MT5ForConditionalGeneration,
    BartForConditionalGeneration,
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    Adafactor,
    AdamWeightDecay
)

from translate_package.models.gradient_observation import get_gradients_mean

from translate_package.models.lstm import LSTMSequenceToSequence

import os


def print_number_of_trainable_model_parameters(model):
    trainable_model_params = 0
    all_model_params = 0
    for _, param in model.named_parameters():
        all_model_params += param.numel()
        if param.requires_grad:
            trainable_model_params += param.numel()
    return f"trainable model parameters: {trainable_model_params}\nall model parameters: {all_model_params}\npercentage of trainable model parameters: {100 * trainable_model_params / all_model_params:.2f}%"

class MachineTranslationTransformer(pl.LightningModule):

    rouge = evaluate.load("rouge")
    bleu = evaluate.load("sacrebleu")

    def __init__(
        self,
        model_name,
        tokenizer,
        model_generation="t5",
        model=None,
        lr=1e-4,
        weight_decay=1e-2,
        num_warmup_steps=0,
        num_training_steps=20000,
        r=32,
        lora_alpha=32,
        lora_dropout=0.05,
        bias="none",
        max_new_tokens=50,
        predict_with_generate=True,
        num_beams=0,
        use_peft=False,
        embedding_size=128,
        num_layers=6,
        hidden_size=128,
        dropout=0.1,
        bidirectional=False,
        length_penalty=1.2,
    ):

        super().__init__()

        if model is None:
            if model_generation in ["t5"]:
                
                self.original_model = T5ForConditionalGeneration.from_pretrained(
                    model_name, torch_dtype=torch.float32
                )
                
            elif model_generation in ["mt5"]:
                
                self.original_model = MT5ForConditionalGeneration.from_pretrained(
                    model_name, torch_dtype=torch.float32
                )
                
            elif model_generation in ["nllb"]:
                
                self.original_model = AutoModelForSeq2SeqLM.from_pretrained(
                    model_name, torch_dtype=torch.float32
                )
                
            elif model_generation in ["bart"]:
                
                self.original_model = BartForConditionalGeneration.from_pretrained(
                    model_name, torch_dtype=torch.float32
                )
            
            elif model_generation in ["lstm"]:
                
                self.original_model = LSTMSequenceToSequence(tokenizer, embedding_size, num_layers, hidden_size, dropout, bidirectional)
                
            # resize the token embeddings
            if not model_generation in ["lstm"]: self.original_model.resize_token_embeddings(len(tokenizer))
            
            if use_peft and not model_generation in ["lstm"]:
                
                self.lora_config = LoraConfig(
                    r=r,  # Rank
                    lora_alpha=lora_alpha,
                    lora_dropout=lora_dropout,
                    bias=bias,
                    task_type=TaskType.SEQ_2_SEQ_LM, 
                )
                
                self.model = get_peft_model(self.original_model, self.lora_config)          
            
            else:
                
                self.model = self.original_model
            
        else:

            self.model = model

        print(print_number_of_trainable_model_parameters(self.model))

        self.tokenizer = tokenizer
        
        self.lr = lr

        self.weight_decay = weight_decay

        self.num_warmup_steps = num_warmup_steps

        self.num_training_steps = num_training_steps

        self.predict_with_generate = predict_with_generate

        self.max_new_tokens = max_new_tokens

        self.num_beams = num_beams
        
        self.length_penalty = length_penalty
        
        self.model_generation = model_generation

        self.predictions = {
            "Source references": [],
            "Predictions": [],
            "Target references": [],
        }

    def forward(self, input):

        output = self.model(**input) if not self.model_generation in ["lstm"] else self.model(input['input_ids'], input['labels'])

        return output.loss, output.logits

    def configure_optimizers(self):

        if self.model_generation in ["t5"]:
            
            optimizer = Adafactor(
                self.parameters(), lr=self.lr, weight_decay=self.weight_decay, relative_step = False,
                warmup_init = False
            )
        
        elif self.model_generation in ["bart", "mbart", "nllb", "mt5"]:
            
            optimizer = torch.optim.AdamW(
                self.parameters(), lr=self.lr, weight_decay=self.weight_decay
            )
        
        elif self.model_generation in ["lstm"]:
            
            optimizer = torch.optim.AdamW(
                self.parameters(), lr=self.lr, weight_decay=self.weight_decay
            )

        if self.model_generation in ["t5", "lstm"]:
            
            return [optimizer]

        elif self.model_generation in ["bart", "nllb", "mt5"]:

            scheduler = get_linear_schedule_with_warmup(
                optimizer,
                num_warmup_steps=self.num_warmup_steps,
                num_training_steps=self.num_training_steps,
            )
            
            return {'optimizer': optimizer, 'lr_scheduler': {"scheduler": scheduler}}
            
    def training_step(self, batch, batch_idx=None):

        loss, y_pred = self(batch)

        self.log_dict(
            {"train_loss": loss, "global_step": float(self.global_step)},
            prog_bar=True,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        
        mean_grad = get_gradients_mean(self.original_model)

        wandb.log({"train_loss": loss, "trainer/global_step": self.global_step, "mean_gradient": mean_grad})

        return loss

    def validation_step(self, batch, batch_idx=None):

        loss, y_pred = self(batch)

        metrics = {}

        if self.predict_with_generate:

            # generate predictions
            predictions = self.model.generate(
                input_ids=batch["input_ids"],
                attention_mask=batch["attention_mask"],
                max_new_tokens=self.max_new_tokens,
            ) if not self.model_generation in ["lstm"] else self.model.generate(
                input=batch["input_ids"],
                max_new_tokens=self.max_new_tokens,
            ) 

            # decode the labels
            predictions = self.tokenizer.batch_decode(
                predictions, skip_special_tokens=True
            )
            labels = self.tokenizer.batch_decode(
                batch["labels"], skip_special_tokens=True
            )

            # get bleu metric
            bleu = self.bleu.compute(
                predictions=predictions,
                references=[[label.strip()] for label in labels],
            )

            metrics["bleu"] = bleu["score"]

            # get rouge metrics
            rouge = self.rouge.compute(
                predictions=predictions, references=[label.strip() for label in labels]
            )

            metrics.update({k: v for k, v in rouge.items() if "rouge" in k})

        # get the loss
        metrics.update(
            {"eval_loss": loss.item(), "global_step": float(self.global_step)}
        )

        self.log_dict(
            metrics, prog_bar=True, on_step=False, on_epoch=True, sync_dist=True
        )

        metrics.update({"trainer/global_step": self.global_step})

        wandb.log(metrics)

        return loss

    def test_step(self, batch, batch_idx):

        loss, y_pred = self(batch)

        references = self.tokenizer.batch_decode(
            batch["input_ids"], skip_special_tokens=True
        )

        # generate predictions
        if not self.model_generation in ["lstm"] and self.num_beams > 0:
            
            predictions = self.model.generate(
                input_ids=batch["input_ids"],
                attention_mask=batch["attention_mask"],
                max_new_tokens=self.max_new_tokens,
                do_sample=True,
                num_beams=self.num_beams,
                length_penalty=self.length_penalty
            )  
            
        elif not self.model_generation in ["lstm"]:
            
            predictions = self.model.generate(
                input_ids=batch["input_ids"],
                attention_mask=batch["attention_mask"],
                max_new_tokens=self.max_new_tokens,
                do_sample=False
            ) 
        
        else:
            
            predictions = self.model.generate(
                input=batch["input_ids"],
                max_new_tokens=self.max_new_tokens,
                use_sampling=True
            )

        # decode the labels
        predictions = self.tokenizer.batch_decode(predictions, skip_special_tokens=True)
        labels = self.tokenizer.batch_decode(batch["labels"], skip_special_tokens=True)

        self.predictions["Source references"].extend(references)
        self.predictions["Predictions"].extend(predictions)
        self.predictions["Target references"].extend(labels)

        # get bleu metric
        bleu = self.bleu.compute(
            predictions=predictions, references=[[label.strip()] for label in labels]
        )

        metrics = {}

        metrics["bleu"] = bleu["score"]

        # get rouge metrics
        rouge = self.rouge.compute(predictions=predictions, references=labels)

        metrics.update({k: v for k, v in rouge.items() if "rouge" in k})

        # get the loss
        metrics.update(
            {"test_loss": loss.item(), "global_step": float(self.global_step)}
        )

        # log metrics
        self.log_dict(
            metrics, prog_bar=True, on_step=False, on_epoch=True, sync_dist=True
        )
        
    def save_model(self, directory: str = "my_model", model_name = "pytorch_model.bin"):
        
        if not os.path.exists(directory):
            
            os.makedirs(directory)
            
        torch.save(self.model.state_dict(), os.path.join(directory, model_name))
        
        if hasattr(self.model, "config"):
            
            self.model.config.to_json_file(os.path.join(directory, "config.json"))
            
        self.tokenizer.save_pretrained(directory)
        
        
        



class EmptyArtifactException(Exception):
    def __init__(
        self, error="No wandb's artifact path was provided!! Please provide one."
    ):

        print(error)
        
class TokenizerException(Exception):
    
    def __init__(self, error: str):
        
        print(error)



import torch

class LSTMOutput:
    
    def __init__(self, logits, loss):
        
        self.logits = logits
        
        self.loss = loss

class LSTMSequenceToSequence(torch.nn.Module):

  def __init__(self, tokenizer, embedding_size = 128, num_layers = 6, hidden_size = 128, dropout=0.1, bidirectional = True):
    
    super().__init__()

    self.tokenizer = tokenizer
    
    self.vocab_size = self.tokenizer.vocab_size

    self.embedding = torch.nn.Embedding(self.vocab_size, embedding_size)

    self.encoder = torch.nn.LSTM(input_size = embedding_size, hidden_size = hidden_size, num_layers = num_layers, batch_first = True,
                                 bidirectional = bidirectional, dropout=dropout)

    self.decoder = torch.nn.LSTM(input_size = embedding_size, hidden_size = hidden_size, num_layers = num_layers, batch_first = True,
                                 bidirectional = bidirectional, dropout=dropout)

    copy = 2 if bidirectional else 1

    self.decoder_output_layer = torch.nn.Linear(copy * hidden_size, self.vocab_size)

    self.loss_fn = torch.nn.CrossEntropyLoss(ignore_index=tokenizer.pad_token_id)

  def forward(self, input, output):
    
    input_embed = self.embedding(input)

    state, hidden = self.encoder(input_embed)

    # decal output for decoder
    decoder_input = output[:, :-1]

    decoder_input = self.embedding(decoder_input)

    decoder_output, _ = self.decoder(decoder_input, hidden)

    decoder_output = self.decoder_output_layer(decoder_output)
    
    loss = self.loss_fn(decoder_output.reshape(-1, decoder_output.shape[-1]), output[:, 1:].reshape(-1))

    return LSTMOutput(decoder_output, loss)

  def generate(self, input, max_new_tokens: int = 100, temperature: float = 0.0, use_sampling = False, **kwargs):

    input_embed = self.embedding(input)

    _, hidden = self.encoder(input_embed)

    # initialize predictions
    predictions = torch.tensor([[self.tokenizer.bos_token_id]]*input_embed.shape[0], dtype=torch.long, device=input.device)
    
    # variable identifying if the sequence is finished
    finished_sequences = torch.zeros(input_embed.shape[0], dtype=torch.bool, device = input.device)
    
    # generate predictions
    for i in range(max_new_tokens):

      decoder_input = self.embedding(predictions)

      decoder_output, hidden = self.decoder(decoder_input, hidden)

      decoder_output = self.decoder_output_layer(decoder_output)

      if temperature > 0.0: decoder_output = (decoder_output / temperature)

      # get probs and sample the next token from a multinomial distribution
      probs = torch.softmax(decoder_output[:, -1], dim = -1)

      if use_sampling: prediction = torch.multinomial(probs, num_samples = 1)
      else: prediction = torch.argmax(probs, dim=-1, keepdim=True)

      # add new prediction
      predictions = torch.cat((predictions, prediction), dim = -1)

      # recuperate next ids
      next_token_ids = prediction.squeeze(-1)
      
      finished_sequences |= (next_token_ids == self.tokenizer.eos_token_id)

      if finished_sequences.all():
        
        break
      
    # return predictions
    return predictions[:, 1:]




from huggingface_hub import login, HfApi, upload_folder, create_repo


def upload_model(hub_token, directory = "my_model", username = "", repo_name = "", commit_message = "new model created"):
    
    repo_id = f"{username}/{repo_name}"
    
    login(token=hub_token)
    
    create_repo(repo_id, token = hub_token)
    
    upload_folder(repo_id = repo_id, folder_path = directory, commit_message= commit_message, token=hub_token)
    
    print(f"Model was successfully upload to {repo_id}.")
    
    



from translate_package import argparse, spm, pd, os, Tokenizer, models, pre_tokenizers, trainers, processors, decoders
from translate_package.errors import TokenizerException

def train_tokenizer(arguments):
    
    # recuperate dataset
    dataset = pd.read_csv(arguments.dataset_file).astype(str)
    
    if arguments.name == 'bpe':
        
        # initialize tokenizer
        tokenizer = Tokenizer(models.BPE())
        
        # initialize pre-to
        tokenizer.pre_tokenizer = pre_tokenizers.ByteLevel(add_prefix_space = False)
        
        # initialize trainer
        trainer = trainers.BpeTrainer(vocab_size = arguments.vocab_size, special_tokens = ['<s>', '<pad>', '</s>', '<unk>'])
        
        # iterate over the dataset and return the sentences
        def get_training_corpus():
            
            sentences = dataset[arguments.src_label].tolist() + dataset[arguments.tgt_label].tolist()
            
            for i in range(0, len(sentences), arguments.batch_size):
                
                yield sentences[i: i + arguments.batch_size]
                
        # train the tokenizer
        tokenizer.train_from_iterator(get_training_corpus(), trainer = trainer)
        
        tokenizer.post_processor = processors.ByteLevel(trim_offsets = False)
        
        tokenizer.decoder = decoders.ByteLevel()
        
        # get path
        tk_path = os.path.join(arguments.save_path, f'{arguments.file_name}.json')
        
        tokenizer.save(tk_path)
        
        print(f"The Byte Pair Encoding tokenizer was saved as {tk_path}!")
    
    elif arguments.name == 'sp':
        
        # print sentences into a file
        with open('sents.txt', 'w', encoding = 'utf-8') as f:
            
            sentences =  dataset[arguments.src_label].tolist() + dataset[arguments.tgt_label].tolist()
            
            for i in range(0, len(sentences), arguments.batch_size):
                
                sents = sentences[i: i + arguments.batch_size]
                
                f.write("\n".join(sents)+'\n')
        
        # get path
        tk_path = os.path.join(arguments.save_path, arguments.file_name)
              
        # initialize sentence piece trainer
        spm.SentencePieceTrainer.Train(input = f'sents.txt',
                               model_prefix=os.path.join(arguments.save_path, arguments.file_name),
                               vocab_size=arguments.vocab_size,
                               character_coverage=1.0,
                               pad_id=0,                
                               eos_id=1,
                               unk_id=2,
                               bos_id=3,
                               pad_piece='<pad>',
                               eos_piece='</s>',
                               unk_piece='<unk>',
                               bos_piece='<s>',
                              )
        
        # remove file
        os.remove('sents.txt')
        
        print(f"The Sentence Piece tokenizer was saved as {tk_path}(.model / for model)!")
        
    else:
        
        raise TokenizerException("You can only train a sentence piece (as 'sp') tokenizer, or a byte pair encoding (as 'bpe') tokenizer!")
    


# -*- coding: utf-8 -*-
from tokenizers import (
    decoders,
    models,
    normalizers,
    pre_tokenizers,
    processors,
    trainers,
    Tokenizer
)
from transformers import (
    GenerationConfig,
    TrainingArguments,
    Trainer, AutoModelForSeq2SeqLM,
    get_linear_schedule_with_warmup,
    T5ForConditionalGeneration, Adafactor, BartForConditionalGeneration,
    MT5ForConditionalGeneration, AdamWeightDecay, AutoTokenizer
)
from wolof_translate.utils.bucket_iterator import SequenceLengthBatchSampler, BucketSampler
from wolof_translate.utils.sent_transformers import TransformerSequences
from wolof_translate.utils.sent_corrections import *
from peft import LoraConfig, get_peft_model, TaskType
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, Dataset
from nlpaug.augmenter import char as nac
import matplotlib.pyplot as plt
import pytorch_lightning as pl
from functools import partial
import sentencepiece as spm
from math import ceil
import pandas as pd
import numpy as np
import argparse
import evaluate
import string
import random
import shutil
import wandb
import torch
import time
import nltk
import os

# Désactiver le parallélisme des tokenizers
os.environ["TOKENIZERS_PARALLELISM"] = "false"
